import re

from depgen import IncludeFinder, StreamStack, file_in_dir, open23, \
    find_unquoted_string


class FortranParser:
    _re_include = re.compile(r'^\s*include\s+(\'|")(.*?)\1', re.I)
    _re_line_continue_start = re.compile(r'^(.*)&\s*$')
    _re_line_continue_end = re.compile(r'^\s*&')
    _re_module_provide = re.compile(r'^\s*module\s+(?!procedure\s)(\w+)', re.I)
    _re_module_require = re.compile(
        r'^\s*use(?:\s+|(?:\s*,\s*((?:non_)?intrinsic))?\s*::\s*)(\w+)', re.I)

    def __init__(self,
                 include_order=None,
                 include_dirs=None,
                 include_roots=None,
                 intrinsic_mods=None,
                 external_mods=None):
        self.include_roots = include_roots

        if intrinsic_mods:
            self.intrinsic_mods = set(intrinsic_mods)
        else:
            self.intrinsic_mods = set()

        if external_mods:
            self.external_mods = set(external_mods)
        else:
            self.external_mods = set()

        # Callbacks:
        self.include_callback = None
        self.module_callback = None
        self.use_module_callback = None
        self.debug_callback = None

        self._include_finder = IncludeFinder(include_order, include_dirs)

    def parse(self, stream):
        with StreamStack() as s:
            s.add(stream)
            while 1:
                line = s.readline()
                if not line:
                    break

                # delete comments
                line = FortranParser._delete_comments(line)
                if line.isspace():
                    continue

                # line continuation
                match = FortranParser._re_line_continue_start.match(line)
                while match:
                    next_line = s.readline()
                    if not next_line:
                        break

                    next_line = FortranParser._delete_comments(next_line)

                    # If the line contains only comments, we need the next one
                    # TODO: implement a separate class FortranPrepcocessor
                    if next_line.isspace():
                        continue

                    line = match.group(1) + re.sub(
                        FortranParser._re_line_continue_end, '', next_line)

                    match = FortranParser._re_line_continue_start.match(line)

                for line in FortranParser._split_semicolons(line):
                    # module provided
                    match = FortranParser._re_module_provide.match(line)
                    if match:
                        module_name = match.group(1).lower()
                        if self.module_callback:
                            self.module_callback(module_name)
                        if self.debug_callback:
                            self.debug_callback(
                                line, 'declared module \'%s\'' % module_name)
                        continue

                    # module required
                    match = FortranParser._re_module_require.match(line)
                    if match:
                        module_nature = match.group(1).lower() \
                            if match.group(1) is not None else ''
                        module_name = match.group(2).lower()
                        if module_nature == 'intrinsic':
                            if self.debug_callback:
                                self.debug_callback(
                                    line, 'ignored module usage (\'%s\' '
                                          'is explicitly intrinsic)'
                                          % module_name)
                        elif (module_name in self.intrinsic_mods and
                              module_nature != 'non_intrinsic'):
                            if self.debug_callback:
                                self.debug_callback(
                                    line, 'ignored module usage (\'%s\' '
                                          'is implicitly intrinsic)'
                                          % module_name)
                        elif module_name in self.external_mods:
                            if self.debug_callback:
                                self.debug_callback(
                                    line, 'ignored module usage (\'%s\' '
                                          'is external)' % module_name)
                        else:
                            if self.use_module_callback:
                                self.use_module_callback(module_name)
                            if self.debug_callback:
                                self.debug_callback(
                                    line, 'used module \'%s\'' % module_name)
                        continue

                    # include statement
                    match = FortranParser._re_include.match(line)
                    if match:
                        filename = match.group(2)
                        filepath = self._include_finder.find(
                            filename,
                            s.root_name,
                            s.current_name)
                        if filepath:
                            if not self.include_roots or any(
                                    [file_in_dir(filepath, d)
                                     for d in self.include_roots]):
                                s.add(open23(filepath, 'r'))
                                if self.include_callback:
                                    self.include_callback(filepath)
                                if self.debug_callback:
                                    self.debug_callback(
                                        line, 'included file \'%s\''
                                              % filepath)
                            elif self.debug_callback:
                                self.debug_callback(
                                    line,
                                    'ignored (file \'%s\' '
                                    'is not in the source roots)' % filepath)
                        elif self.debug_callback:
                            self.debug_callback(line,
                                                'ignored (file not found)')
                        continue

    @staticmethod
    def _split_semicolons(line):
        while 1:
            idx = find_unquoted_string(';', line)
            if idx < 0:
                if line and not line.isspace():
                    yield line
                break
            else:
                prefix = line[:idx]
                if prefix and not prefix.isspace():
                    yield prefix + '\n'
                line = line[idx + 1:]

    @staticmethod
    def _delete_comments(line):
        comment_idx = find_unquoted_string('!', line)
        if comment_idx >= 0:
            line = line[:comment_idx]
        return line




import re

from depgen import IncludeFinder, StreamStack, file_in_dir, open23, \
    find_unquoted_string


class Preprocessor:
    _re_ifdef = re.compile(r'^\s*#\s*if(n)?def\s+([a-zA-Z_]\w*)')
    _re_if_expr = re.compile(r'^\s*#\s*if((?:\s|\().*)')

    _re_elif = re.compile(r'^\s*#\s*elif((?:\s|\().*)')
    _re_else = re.compile(r'^\s*#\s*else(?:\s.*)')
    _re_endif = re.compile(r'^\s*#\s*endif(?:\s.*)')

    _re_include = re.compile(r'^\s*#\s*include\s+(?:"(.*?)"|<(.*?)>)')
    _re_define = re.compile(r'^\s*#\s*define\s+([a-zA-Z_]\w*)(\(.*\))?\s+(.*)$')
    _re_undef = re.compile(r'^\s*#\s*undef\s+([a-zA-Z_]\w*)')

    # matches "defined MACRO_NAME" and "defined (MACRO_NAME)"
    _re_defined_call = re.compile(
        r'(defined\s*(\(\s*)?([a-zA-Z_]\w*)(?(2)\s*\)))')

    _re_identifier = re.compile(
        r'(([a-zA-Z_]\w*)(\s*\(\s*(\w+(?:\s*,\s*\w+)*)?\s*\))?)')

    def __init__(self,
                 stream,
                 include_order=None,
                 include_sys_order=None,
                 include_dirs=None,
                 include_roots=None,
                 try_eval_expr=False,
                 inc_sys=False,
                 predefined_macros=None):
        self.include_roots = include_roots
        self.try_eval_expr = try_eval_expr
        self.inc_sys = inc_sys

        # Callbacks:
        self.include_callback = None
        self.debug_callback = None

        self._include_finder = IncludeFinder(include_order, include_dirs)
        self._include_sys_finder = IncludeFinder(include_sys_order,
                                                 include_dirs)
        self._include_stack = StreamStack()
        self._include_stack.add(stream)

        if predefined_macros:
            self._macros = dict(predefined_macros)
        else:
            self._macros = dict()

        # Stack of #if-#else blocks holds one of the following:
        # 1 - keep current branch and ignore another
        # -1 - ignore current branch and keep another
        # 0 - keep both branches (if failed to evaluate expression)
        self._if_state_stack = []

        # Each #elif is interpreted as a combination of #else and #if.
        # Thus, each #elif increments the number of #if blocks that are
        # closed with #endif statement. This numbers are stored in a separate
        # stack:
        self._states_per_endif_stack = []

    def readline(self):
        while 1:
            line = self._include_stack.readline()

            if not line:
                return line

            line = self._replace_continuation(line)
            line = self._remove_block_comments(line)

            if line.isspace():
                continue

            # if(n)def directive
            match = Preprocessor._re_ifdef.match(line)
            if match:
                macro, negate, state = match.group(2), bool(match.group(1)), 0
                if not self._branch_is_dead():
                    state = 1 if bool(macro in self._macros) ^ negate else -1
                    if self.debug_callback:
                        self.debug_callback(
                            line, 'evaluated to {0}'.format(state > 0))
                elif self.debug_callback:
                    self.debug_callback(
                        line, 'was not evaluated (dead branch)')
                self._if(state)
                continue

            # if directive
            match = Preprocessor._re_if_expr.match(line)
            if match:
                expr, state = match.group(1), 0
                if not self._branch_is_dead():
                    if self.try_eval_expr:
                        state = self._evaluate_expr_to_state(expr)
                        if self.debug_callback:
                            self.debug_callback(
                                line, 'evaluated to {0}'.format(
                                    'Unknown (evaluation failed)'
                                    if state == 0 else state > 0))
                    elif self.debug_callback:
                        self.debug_callback(
                            line, 'was not evaluated (evaluation disabled)')
                elif self.debug_callback:
                    self.debug_callback(
                        line, 'was not evaluated (dead branch)')
                self._if(state)
                continue

            # elif directive
            match = Preprocessor._re_elif.match(line)
            if match:
                self._else()
                expr, state = match.group(1), 0
                if not self._branch_is_dead():
                    if self.try_eval_expr:
                        state = self._evaluate_expr_to_state(expr)
                        if self.debug_callback:
                            self.debug_callback(
                                line, 'evaluated to {0}'.format(
                                    'Unknown (evaluation failed)'
                                    if state == 0 else state > 0))
                    elif self.debug_callback:
                        self.debug_callback(
                            line, 'was not evaluated (evaluation disabled)')
                elif self.debug_callback:
                    self.debug_callback(
                        line, 'was not evaluated (dead branch)')
                self._elif(state)
                continue

            # else directive
            match = Preprocessor._re_else.match(line)
            if match:
                self._else()
                continue

            # endif directive
            match = Preprocessor._re_endif.match(line)
            if match:
                self._endif()
                continue

            if self._branch_is_dead() and self.debug_callback is None:
                continue

            # define directive
            match = Preprocessor._re_define.match(line)
            if match:
                if not self._branch_is_dead():
                    self._define(*match.group(1, 2, 3))
                    if self.debug_callback:
                        self.debug_callback(line, 'accepted')
                elif self.debug_callback:
                    self.debug_callback(line, 'ignored (dead branch)')
                continue

            # undef directive
            match = Preprocessor._re_undef.match(line)
            if match:
                if not self._branch_is_dead():
                    self._macros.pop(match.group(1), None)
                    if self.debug_callback:
                        self.debug_callback(line, 'accepted')
                elif self.debug_callback:
                    self.debug_callback(line, 'ignored (dead branch)')
                continue

            # include directive
            match = Preprocessor._re_include.match(line)
            if match:
                if not self._branch_is_dead():

                    if match.lastindex == 1:  # quoted form
                        filepath = self._include_finder.find(
                            match.group(1),
                            self._include_stack.root_name,
                            self._include_stack.current_name)
                    elif match.lastindex == 2:  # angle-bracket form
                        if self.inc_sys:
                            filepath = self._include_sys_finder.find(
                                match.group(2),
                                self._include_stack.root_name,
                                self._include_stack.current_name)
                        else:
                            if self.debug_callback:
                                self.debug_callback(line,
                                                    'ignored (system header)')
                            continue
                    else:
                        if self.debug_callback:
                            self.debug_callback(line,
                                                'ignored (internal error)')
                        continue

                    if filepath:
                        if not self.include_roots or any(
                                [file_in_dir(filepath, d)
                                 for d in self.include_roots]):
                            self._include_stack.add(open23(filepath, 'r'))
                            if self.include_callback:
                                self.include_callback(filepath)
                            if self.debug_callback:
                                self.debug_callback(
                                    line,
                                    "included file '{0}'".format(filepath))
                        elif self.debug_callback:
                            self.debug_callback(
                                line,
                                "ignored (file '{0}' "
                                "is not in the source roots)".format(filepath))
                    elif self.debug_callback:
                        self.debug_callback(line, 'ignored (file not found)')
                elif self.debug_callback:
                    self.debug_callback(line, 'ignored (dead branch)')
                continue

            if self._branch_is_dead():
                continue

            return line

    @property
    def name(self):
        return self._include_stack.root_name

    def close(self):
        self._include_stack.clear()

    def _define(self, name, args=None, body=None):
        if name != 'defined':
            self._macros[name] = (args, '' if body is None else body)

    def _if(self, state):
        self._if_state_stack.append(state)
        self._states_per_endif_stack.append(1)

    def _else(self):
        if self._if_state_stack:
            self._if_state_stack[-1] = -self._if_state_stack[-1]

    def _elif(self, state):
        self._if_state_stack.append(state)
        self._states_per_endif_stack[-1] += 1

    def _endif(self):
        if self._if_state_stack:
            pop_count = self._states_per_endif_stack.pop()
            for _ in range(pop_count):
                self._if_state_stack.pop()

    def _branch_is_dead(self):
        return any(state < 0 for state in self._if_state_stack)

    def _replace_continuation(self, line):
        while line.endswith('\\\n'):
            suffix = self._include_stack.readline()
            line = line[:-2] + suffix
        return line

    def _remove_block_comments(self, line):
        while 1:
            # Check whether the line contains an unquoted block comment
            # initiator '/*':
            start_idx = find_unquoted_string('/*', line)
            if start_idx < 0:
                return line
            # Check whether the line contains a block comment
            # terminator '*/' (even if it is quoted):
            term_idx = line.find('*/', start_idx + 2)
            while term_idx < 0:
                # The block is not terminated yet, read the next line:
                next_line = self._include_stack.readline()
                line_length = len(line)
                if next_line:
                    line += next_line
                    term_idx = line.find('*/', line_length)
                else:
                    term_idx = line_length
            else:
                # Replace the block of comments with a single
                # space:
                line = '%s %s' % (line[:start_idx], line[term_idx + 2:])

    def _evaluate_expr_to_state(self, expr):
        prev_expr = expr
        while 1:
            # replace calls to function "defined"
            defined_calls = re.findall(Preprocessor._re_defined_call, expr)
            for call in defined_calls:
                expr = expr.replace(
                    call[0], '1' if call[2] in self._macros else '0')

            identifiers = re.findall(Preprocessor._re_identifier, expr)

            for identifier in identifiers:
                if identifier[1] == 'defined':
                    return 0

                macro = self._macros.get(identifier[1], None)
                if identifier[2]:
                    # potential call to a function
                    if macro is None:
                        # call to undefined function
                        return 0
                    elif macro[0] is not None:
                        # we can't evaluate function-like macros
                        return 0
                    else:
                        # identifier is defined as object-like macro
                        expr = expr.replace(identifier[0],
                                            macro[1] + identifier[2])
                else:
                    # no function call
                    if macro is None or macro[0] is not None:
                        # macro is not defined or
                        # defined as function-like macro
                        expr = expr.replace(identifier[0], '0')
                    else:
                        # identifier is defined as object-like macro
                        expr = expr.replace(identifier[0], macro[1])

            if prev_expr == expr:
                break
            else:
                prev_expr = expr

        expr = expr.replace('||', ' or ')
        expr = expr.replace('&&', ' and ')
        expr = expr.replace('!', 'not ')

        try:
            result = bool(eval(expr, {}))
            return 1 if result else -1
        except:
            return 0




from netCDF4 import Dataset
import numpy as np
import matplotlib.pyplot as plt

#% This python scripts plots some of the inputs to the radiation scheme
#% in Fig. 1 and the outputs in Fig. 2
code = 'ecrad_meridian'
in_file=Dataset(code+'.nc','r')
pressure_hl=in_file.variables['pressure_hl'][:]
temperature_hl=in_file.variables['temperature_hl'][:]
lat=in_file.variables['lat'][:]
cos_solar_zenith_angle=in_file.variables['cos_solar_zenith_angle'][:]
skin_temperature=in_file.variables['skin_temperature'][:]
cloud_fraction=in_file.variables['cloud_fraction'][:]
aerosol_mmr=in_file.variables['aerosol_mmr'][:]
cases = [code+'_noaer_out.nc',
	 code+'_default_out.nc',
	 code+'_expran_out.nc',
	 code+'_tc_out.nc',
	 code+'_spartacus_out.nc',
	 code+'_spartacus_maxentr_out.nc']


f = [0]*6
k=0
for icase in cases:
    f[k] = Dataset(icase,'r')
    k=k+1



case_list = [0,1,2,3,4,5]
leg = ['McICA Exp-Exp no aerosols',
       'McICA Exp-Exp',
       'McICA Exp-Ran',
       'Tripleclouds Exp-Ran',
       'SPARTACUS Exp-Ran',
       'Classic SPARTACUS Exp-Ran']

styles = ['b','r','g','m','c','k']


p = 0.01*0.5*np.median(pressure_hl[:,0:-1]+pressure_hl[:,1::],0)

nplot=4
plt.figure(1,figsize=(10,10))
# set(gcf,'defaultlinelinewidth',1);
# set(gcf,'units','inches','paperposition',[0.5 0.5 15 30]);
# nplot = 4;
#fig, (ax0, ax1) = plt.subplots(nrows=2)

plt.subplot(nplot,1,1)
plt.plot(lat,cos_solar_zenith_angle,color='k', linewidth=2.0)
ax= plt.gca()
ax.set_xlim([-90,90])
plt.ylabel('Cos solar zenith ang.')


plt.subplot(nplot,1,2)
plt.plot(lat,skin_temperature-273.15,color='r',linewidth=2.0,label='Skin temperature')
plt.plot(lat,temperature_hl[:,-1]-273.15,color='b', linewidth=2.0,label='Air between first two model levels')
plt.gca().set_xlim([-90,90])
plt.ylabel('Temperature ($^\circ$C)');
plt.gca().set_ylim([-50,60]);
plt.gca().legend(loc='best',fontsize='xx-small')

plt.subplot(nplot,1,3)
CS=plt.contourf(lat,p,cloud_fraction.T,10, vmin=0.05, vmax=0.95,cmap="viridis")
plt.colorbar()
plt.contour(CS,levels=CS.levels[::4],colors='k')
plt.gca().set_ylim([0,1013])
plt.gca().invert_yaxis()
plt.gca().set_xlim([-90,90])
plt.ylabel('Pressure (hPa)')
textstyle = dict(size=10, color='white')
plt.gca().text(-70,100,'Cloud fraction',**textstyle)
# text(-90, 0, [' ' 10 '  Cloud fraction'],'verticalalignment','top')


print(aerosol_mmr.shape)
plt.subplot(nplot,1,4)
sea_salt = 1e9*np.sum(aerosol_mmr[:,0:2,-1],1);
dust = 1e9*np.sum(aerosol_mmr[:,3:5,-1],1);
organics = 1e9*np.sum(aerosol_mmr[:,6:7,-1],1);
black_carbon = 1e9*np.sum(aerosol_mmr[:,8:9,-1],1);
sulphate = 1e9*np.sum(aerosol_mmr[:,10:11,-1],1);

plt.semilogy(lat,sea_salt,'b',label='Sea salt')
plt.semilogy(lat,dust,'r',label='Dust')
plt.semilogy(lat,organics,'g',label='Organics')
plt.semilogy(lat,black_carbon,'k',label='Black carbon')
plt.semilogy(lat,sulphate,'m',label='Sulphate')
plt.gca().legend(loc='best',fontsize='xx-small')
plt.ylabel('Aerosol mass mixing ratio\n ($\mu$g/kg)')
plt.xlabel('Latitude (deg N)')
plt.gca().set_xlim([-90,90])

plt.savefig('input.png')

nplot=4
plt.figure(2,figsize=(10,10))
# set(gcf,'defaultlinelinewidth',1);
# set(gcf,'units','inches','paperposition',[0.5 0.5 15 30]);

plt.subplot(nplot,1,1)
for icase in case_list:
    print(leg[icase])
    plt.plot(lat, f[icase].variables['flux_dn_sw'][:,-1],label=leg[icase])
    plt.gca().set_xlim([-90, 90]);
    plt.ylabel('Surface SW down \n(W m$^{-2}$)');
    plt.gca().legend(loc='best',fontsize='xx-small')


plt.subplot(nplot,1,2)
for icase in case_list:
    cre =  f[icase].variables['flux_dn_sw'][:,1]-f[icase].variables['flux_dn_sw_clear'][:,1]-f[icase].variables['flux_up_sw'][:,1]+f[icase].variables['flux_up_sw_clear'][:,1]
    plt.plot(lat, cre,label=leg[icase])
    plt.gca().set_xlim([-90, 90]);
    plt.ylabel('SW cloud radiative effect \n(W m$^{-2}$)');
    plt.gca().legend(loc='best',fontsize='xx-small')


plt.subplot(nplot,1,3)
for icase in case_list:
    cre =  f[icase].variables['flux_dn_lw'][:,1]-f[icase].variables['flux_dn_lw_clear'][:,1]-f[icase].variables['flux_up_lw'][:,1]+f[icase].variables['flux_up_lw_clear'][:,1]
    plt.plot(lat, cre,label=leg[icase])
    plt.gca().set_xlim([-90, 90]);
    plt.ylabel('LW cloud radiative effect \n(W m$^{-2}$)');
    plt.gca().legend(loc='best',fontsize='xx-small')

plt.subplot(nplot,1,4)
for icase in case_list:
    plt.plot(lat, f[icase].variables['cloud_cover_sw'],label=leg[icase]);
    plt.gca().set_xlim([-90, 90]);
    plt.ylabel('Cloud cover');
    plt.gca().legend(loc='best',fontsize='xx-small')
    plt.xlabel('Latitude (deg N)');


plt.savefig('output.png')
#plt.show()



# (c) British Crown Copyright 2020, the Met Office.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#   * Neither the name of the Met Office nor the names of its contributors may
#     be used to endorse or promote products derived from this softwarewithout
#     specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.

import netCDF4,argparse,sys
import numpy as np

def get_var_list(ncfile):
    """
    Returns a list with the names of the variables in a NetCDF file.
    """
    dataset = netCDF4.Dataset(ncfile)
    var_list = dataset.variables.keys()
    dataset.close()
    return var_list
    

def read_var(fname, vname):
    """
    Reads a variable from a NetCDF file.
    
    Arguments:
        fname: path to NetCDF file.
        vname: variable name.
    """
    f_id = netCDF4.Dataset(fname,'r')
    return f_id.variables[vname][:]

def calculate_stats(tst, kgo, atol=0.0, rtol=None):
    """
    Returns a dictionary with some basic summary statistics of the differences
    between two numpy arrays. When a threshold for relative differences is
    passed, then the results refer to relative differences.
    The dictionary contains the following keys:
        'N': number of differences.
        'AvgDiff': average difference.
        'MinDiff': minimum difference.
        'MaxDiff': maximum difference.
        'StDev': standard deviation of the differences.
    
    Arguments:
        tst: test variable (numpy array).
        kgo: reference variable (numpy array).
    Keyword arguments:
        atol: absolute tolerance threshold. Smaller differences pass the test.
        rtol: relative tolerance threshold. Smaller differences pass the test.
    """
    summary_stats = {'N':0, 'AvgDiff':0.0, 'MinDiff':0.0, 'MaxDiff':0.0, 'StDev':0.0}
    # All differences
    try:
        d = tst - kgo
    except:
        print("Error: arrays have different shapes.")
        d = kgo
    # Mask for differences larger than absolute tolerance
    maskAllDiff = (np.absolute(d) > atol)
    NallDiff = maskAllDiff.sum()
    # If there are differences larger than atol,
    # then calculate summary statsitics
    if (NallDiff > 0):
        diffs = d[maskAllDiff]
        # Are relative differences requested?
        if rtol is not None:
            # Calculate relative differences. When KGO=0,
            # use test as reference (i.e. relative diff will
            # be 1 or -1)
            maskedKgo = kgo[maskAllDiff]
            rdiffs = diffs / maskedKgo
            rdiffs[maskedKgo == 0.0] = np.sign(diffs[maskedKgo == 0.0])
            # Keep only those diffs larger than relative tolerance
            diffs = rdiffs[np.absolute(rdiffs) > rtol]
            NallDiff = len(diffs)
        # Calculate summary stats
        summary_stats['N'] = NallDiff
        if NallDiff > 0:
            summary_stats['AvgDiff'] = diffs.mean()
            summary_stats['MinDiff'] = diffs.min()
            summary_stats['MaxDiff'] = diffs.max()
            summary_stats['StDev'] = diffs.std()
        
    return summary_stats

def print_stats_table(summary_stats, print_all=False, stats_file=None):
    """
    Print table of summary statistics.
    
    Arguments:
      summary_stats: dictionary with summary statistics. Output from call
                     to function calculate_stats.

    Keywords:
      print_all: by default, it only prints lines with differences,
                 unless print_all==True.
      stats_file: full path to output file.
    """
    # Open file if required
    if stats_file is not None: 
        stdout = sys.stdout
        f_id = open(stats_file, 'w')
        sys.stdout = f_id
    # Header and column names
    print(42*'=', ' Summary statistics ', 42*'=')
    line = ('{:>40s} {:>10s} {:>12s} {:>12s} {:>12s} {:>12s}').format('Variable',
            'N', 'AvgDiff', 'MinDiff', 'MaxDiff', 'StDev')
    print(line)
    # Main table.
    for vname, vstats in summary_stats.items():
        line = ('{vname:>40s} {N:10d} {AvgDiff:12.4e} {MinDiff:12.4e} '
                '{MaxDiff:12.4e} {StDev:12.4e}').format(vname=vname,**vstats)
        if (vstats['N'] > 0):
            print(line)
        else:
            if print_all: print(line)
    print(106*'=')
    # Close file and restore stdout
    if stats_file is not None:
        f_id.close()
        sys.stdout = stdout

#######################
# Main
#######################
if __name__ == '__main__':

    # Command line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("kgo_file", help="File with known good outputs.")
    parser.add_argument("tst_file", help="Test output file.")
    parser.add_argument("--atol", type=float, default=0.0,
                        help="Absolute tolerance.")
    parser.add_argument("--rtol", type=float, default=None,
                        help="Relative tolerance.")
    parser.add_argument("--allvar", type=bool, default=False,
                        help="Print summary stats for all variables.")
    parser.add_argument("--noerror", type=bool, default=False,
                        help="Exit with no error even if differences are "
                             "bigger than tolerances.")
    parser.add_argument("--stats_file", default=None,
                        help="Output file for summary statistics.")
    args = parser.parse_args()

    # Get list of variables
    kgo_vars = get_var_list(args.kgo_file)
    tst_vars = get_var_list(args.tst_file)
    nkgo = len(kgo_vars)
    ntst = len(tst_vars)

    # Colours for error messages
    red_colour   = '\033[91m'
    green_colour = '\033[92m'
    std_colour   = '\033[0m'

    # Dictionary for summary statistics
    summary_stats = {}

    # Iterate over shortest list and calculate stats
    errored = False
    if (nkgo <= ntst):
        vlst = kgo_vars
    else:
        vlst = tst_vars
    for vname in vlst:
        kgo = read_var(args.kgo_file, vname) # KGO
        tst = read_var(args.tst_file, vname) # test
        summary_stats[vname] = calculate_stats(tst, kgo, 
                                               atol=args.atol, rtol=args.rtol)
        if summary_stats[vname]['N'] > 0: errored = True

    # Print summary stats
    if errored:
        print(red_colour + "===== ERROR: some of the differences are larger "
              "than the tolerances." + std_colour)
        print_stats_table(summary_stats, print_all=args.allvar)
    print_stats_table(summary_stats, print_all=args.allvar,
                      stats_file=args.stats_file)

    # Check if large differences should be treated as errors
    if args.noerror: errored = False

    # Error if files have different number variables. If the number 
    # of variables is the same but they have different names, it will
    # fail in summary_stats.
    if (nkgo != ntst):
        errored = True
        print(red_colour +
              "===== ERROR: files differ in the number of variables." +
              std_colour)
        print("===== Variables in KGO: ", nkgo)
        print(kgo_vars)
        print("===== Variables in Test: ", ntst)
        print(tst_vars)

    # Exit with correct error condition
    if errored:
        print(red_colour + "===== ERROR: the test is exiting with an error, "
              "please review the output above." + std_colour)
        sys.exit(1)
    else:
        print(green_colour + "===== Test completed successfully: "
              "all outputs are within the expected tolerances." + std_colour)
        sys.exit()
